unit TileImage;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls,StdCtrls;

type
  TTileImage = class(TPaintBox)
  private
    FTileImage:Boolean;
    FPicture: TPicture;
    XDrawing: Boolean;
    procedure PictureChanged(Sender: TObject);
    procedure SetPicture(Value: TPicture);
    procedure SetTileImage (AValue:Boolean);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DrawToCanvas (ACanvas:TCanvas;ALeft,ATop,AWidth,AHeight:Integer);
  published
    property DragCursor;
    property DragMode;
    property ParentShowHint;
    property Picture: TPicture read FPicture write SetPicture;
    property PopupMenu;
    property ShowHint;
    property TileImage:Boolean read FTileImage write SetTileImage;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure TTileImage.SetTileImage (AValue:Boolean);
begin
  If FTileImage<>AValue then
  begin
    FTileImage:=AValue;
    Invalidate;
  end;
end;

constructor TTileImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FTileImage:=False;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  Height := 105;
  Width := 105;
end;

destructor TTileImage.Destroy;
begin
  FPicture.Free;
  inherited Destroy;
end;

procedure TTileImage.Paint;
var
  Save: Boolean;
  w,h,x,y:Integer;
  cw,ch:Integer;
begin
  if csDesigning in ComponentState then
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
  Save := XDrawing;
  XDrawing := True;
  try
    with inherited Canvas do
    begin
      If (FPicture.Graphic<>nil) then
      begin
        If FTileImage then
        begin
          w:=Picture.Width;
          h:=Picture.Height;
          y:=0;
          cw:=Width;
          ch:=Height;
          while y<ch do
          begin
            x:=0;
            while x<cw do
            begin
              Draw (x,y,Picture.Graphic);
              Inc (x,w);
            end;
            Inc (y,h);
          end;
        end
        else
          Draw(0,0, Picture.Graphic);
      end;
    end;
  finally
    XDrawing := Save;
  end;
end;

procedure TTileImage.DrawToCanvas (ACanvas:TCanvas;ALeft,ATop,AWidth,AHeight:Integer);
var w,h,x,y:Integer;
begin
  with ACanvas do
  begin
    w:=FPicture.Width;
    h:=FPicture.Height;
    y:=Top-ATop;
    while y<AHeight do
    begin
      If y+h>=0 then
      begin
        x:=Left-ALeft;
        while x<AWidth do
        begin
          If x+w>=0 then
            ACanvas.Draw (x,y,FPicture.Graphic);
          Inc (x,w);
        end;
      end;
      Inc (y,h);
    end;
  end;
end;

procedure TTileImage.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
  Invalidate;
end;

procedure TTileImage.PictureChanged(Sender: TObject);
begin
  if not XDrawing then Invalidate;
end;

procedure Register;
begin
  RegisterComponents('OpenTTD', [TTileImage]);
end;

end.
