unit uCrop;
{************************************************************************
 * This program is  free software; you can redistribute it  and modify it
 * under the terms of the GNU  General Public License as published by the
 * Free Software Foundation; either version 2 of the license  or (at your
 * option) any later version.
 ***********************************************************************}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, pngimage, Spin, TileImage, GIFImage;

type
  TfrmCropSprite = class(TForm)
    ScrollBox1: TScrollBox;
    btnCrop: TButton;
    sprite: TImage;
    btnCancel: TButton;
    btnSave: TButton;
    cropBorder_left: TImage;
    cropBorder_top: TImage;
    cropBorder_right: TImage;
    cropBorder_bottom: TImage;
    cropVisual: TShape;
    Label3: TLabel;
    cropTop: TSpinEdit;
    Label2: TLabel;
    cropLeft: TSpinEdit;
    Label4: TLabel;
    cropBottom: TSpinEdit;
    cropRight: TSpinEdit;
    Label1: TLabel;
    TileImage1: TTileImage;
    spriteOutline: TShape;
    procedure SetInitialCropSettings(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure UpdateCropSettings(Sender: TObject);
    procedure btnCropClick(Sender: TObject);
  private
    { Private declarations }
    loader:boolean;
  public
    { Public declarations }

    procedure SetUp();
    procedure ReSetUp();

    function DetectEmptyParts():TRect;

    procedure UpdateVisualCrop();

  end;

var
  frmCropSprite: TfrmCropSprite;

implementation

{$R *.dfm}

uses uMain;

{$R crop_bg.res}

procedure TfrmCropSprite.SetUp;
var
  RS: TResourceStream;
  png:TpngImage;
begin
// set sprite position
sprite.Left:= 0;
sprite.Top := 0;
spriteOutline.Left :=0;
spriteOutline.Top :=0;
spriteOutline.Width := sprite.Width;
spriteOutline.Height:= sprite.Height;
// get glass settings from main form
if frmSpriteEditor.GlassFrame.Enabled then
  begin
  GlassFrame.Enabled := true;
  GlassFrame.SheetOfGlass := true;
  ScrollBox1.Color := clBlack;
  TileImage1.Visible := false;
  spriteOutline.Visible := true;
  end
else
  begin
  TileImage1.Visible := true;
  TileImage1.TileImage := true;
  TileImage1.Width := sprite.Width;
  TileImage1.Height := sprite.Height;
  spriteOutline.Visible := false;
  end;

// initiate crop visuals
cropVisual.Visible := false;
cropBorder_left.Visible := false;
cropBorder_top.Visible := false;
cropBorder_right.Visible := false;
cropBorder_bottom.Visible := false;

cropVisual.Left:=0;
cropVisual.Top:=0;
cropVisual.Width := sprite.Width;
cropVisual.Height:= sprite.Height;

RS := TResourceStream.Create(0,'CROP_BG','PNG');
RS.Position := 0;
png := TPNGImage.Create;
png.LoadFromStream(RS);

cropBorder_left.Picture.Assign(png);
cropBorder_top.Picture.Assign(png);
cropBorder_right.Picture.Assign(png);
cropBorder_bottom.Picture.Assign(png);
RS.Free;
png.Free;
end;

{ this function is called after cropping}
procedure TfrmCropSprite.ReSetUp;
begin
// set sprite position
sprite.Left:= 0;
sprite.Top := 0;
spriteOutline.Left :=0;
spriteOutline.Top :=0;
spriteOutline.Width := sprite.Width;
spriteOutline.Height:= sprite.Height;

TileImage1.Width := sprite.Width;
TileImage1.Height := sprite.Height;

cropLeft.Value := 0;
cropTop.Value  := 0;
cropRight.Value  := 0;
cropBottom.Value := 0;

// initiate crop visuals
cropVisual.Visible := false;
cropBorder_left.Visible := false;
cropBorder_top.Visible := false;
cropBorder_right.Visible := false;
cropBorder_bottom.Visible := false;

cropVisual.Left:=0;
cropVisual.Top:=0;
cropVisual.Width := sprite.Width;
cropVisual.Height:= sprite.Height;

end;

function TfrmCropSprite.DetectEmptyParts():TRect;
var png:TPngImage;
    alpha: byte;
    pixel :Tcolor;
    x,y,x2,y2,x3,y3:integer;

    crop:TRect;
begin

if not (sprite.Picture.Graphic is TPNGImage) then Result:=Rect(0,0,0,0)
else
  begin
  png:= sprite.Picture.Graphic as TPngImage;

  crop.Left := png.Header.Width-1;
  crop.Top := png.Header.Height-1;
  crop.Right := 0;
  crop.Bottom := 0;

  for y:= 0 to png.Header.Height-1 do
    begin
    x2 := crop.Left;  y2 := crop.Top;
    x3 := crop.Right; y3 := crop.Bottom;

    for x := 0 to png.Header.Width - 1 do
      begin
      pixel:= png.Canvas.Pixels[x,y];
      alpha:= png.AlphaScanline[y]^[x];

     if alpha <> 0 then
        begin
        if (x < x2) then x2 := x;
        if (x >x3) then x3 := x;

        if (y < y2) then y2 := y;
        if (y > y3) then y3 := y;
        end;
      end;
    if x2 < crop.Left then crop.Left := x2;
    if x3 > crop.Right then crop.Right := x3;

    if y2 < crop.Top then crop.Top := y2;
    if y3 > crop.Bottom then crop.Bottom := y3;

    end;

  crop.Right := crop.Right+1;
  crop.Bottom := crop.Bottom+1;

  if crop.Left >= png.Header.Width then crop.Left := 0;
  if crop.Top >= png.Header.Height then crop.Top := 0;

  Result := crop;
  end;
end;

procedure TfrmCropSprite.UpdateVisualCrop();
begin
if (cropLeft.Value > 0) or (cropTop.Value > 0) or (cropRight.Value > 0) or (cropBottom.Value > 0) then
  begin
  cropVisual.Left := cropLeft.Value;
  cropVisual.Top := cropTop.Value;
  cropVisual.Width := sprite.Width - cropLeft.Value - cropRight.Value;
  cropVisual.Height:= sprite.Height- cropTop.Value - cropBottom.Value;
  cropVisual.Visible := true;
  if cropLeft.Value > 0 then
    begin
    cropBorder_left.Left:= 0;
    cropBorder_left.Top := 0;
    cropBorder_left.Width := cropVisual.Left;
    cropBorder_left.Height := sprite.Height;
    cropBorder_left.Visible := true;
    end
  else cropBorder_left.Visible := false;

  if cropTop.Value > 0 then
    begin
    cropBorder_top.Left:= cropVisual.Left;
    cropBorder_top.Top := 0;
    cropBorder_top.Width := cropVisual.Width;
    cropBorder_top.Height := cropVisual.Top;
    cropBorder_top.Visible := true;
    end
  else cropBorder_top.Visible := false;

  if cropRight.Value > 0 then
    begin
    cropBorder_right.Left:= cropVisual.Left+cropVisual.Width;
    cropBorder_right.Top := 0;
    cropBorder_right.Width := sprite.Width - cropVisual.Width - cropVisual.Left;
    cropBorder_right.Height := sprite.Height;
    cropBorder_right.Visible := true;
    end
  else cropBorder_right.Visible := false;

  if cropBottom.Value > 0 then
    begin
    cropBorder_bottom.Left:= cropVisual.Left;
    cropBorder_bottom.Top := cropVisual.Top+cropVisual.Height;
    cropBorder_bottom.Width := cropVisual.Width;
    cropBorder_bottom.Height := sprite.Height - cropVisual.Height - cropVisual.Top;
    cropBorder_bottom.Visible := true;
    end
  else cropBorder_bottom.Visible := false;
  end
else
  begin
    // no values above zero
  cropVisual.Visible := false;
  cropBorder_left.Visible := false;
  cropBorder_top.Visible := false;
  cropBorder_right.Visible := false;
  cropBorder_bottom.Visible := false;
  end;
end;













procedure TfrmCropSprite.FormCreate(Sender: TObject);
begin
loader := false;
end;

procedure TfrmCropSprite.SetInitialCropSettings(Sender: TObject);
var coord:TRect;
begin
coord := DetectEmptyParts();

loader := true;

cropLeft.Value := coord.Left;
cropTop.Value  := coord.Top;
cropRight.Value  := sprite.Width - coord.Right;
cropBottom.Value := sprite.Height - coord.Bottom;

loader := false;

UpdateVisualCrop();

end;

procedure TfrmCropSprite.UpdateCropSettings(Sender: TObject);
begin
if not loader then
  begin
  if (Sender as TSpinEdit).Value < 0 then (Sender as TSpinEdit).Value := 0;
  UpdateVisualCrop();
  end;
end;

procedure TfrmCropSprite.btnCropClick(Sender: TObject);
var png,png2:TPngImage;
    alpha: byte;
    pixel :Tcolor;
    x,y,x2,y2,x3,y3:integer;

    crop,c2:TRect;
begin
if not (sprite.Picture.Graphic is TPNGImage) then exit;

loader := true;

png:= sprite.Picture.Graphic as TPngImage;

crop.Left := cropLeft.Value;
crop.Top := cropTop.Value;
crop.Right := png.Header.Width - cropRight.Value;
crop.Bottom:= png.Header.Height -cropBottom.Value;

png2 := TPngImage.CreateBlank(COLOR_RGB, 16, cropVisual.Width, cropVisual.Height);
png2.CreateAlpha;

//png.Draw(png2.canvas, crop);

// fix transparency
for y := 0 to png2.Header.Height - 1 do
  for x := 0 to png2.Header.Width - 1 do
    png2.AlphaScanline[y]^[x] := png.AlphaScanline[crop.Top+y]^[crop.Left+x];

c2.Left:=0;
c2.Top:=0;
c2.Right :=  png2.Header.Width;
c2.Bottom := png2.Header.Height;

png2.canvas.CopyRect(c2,png.Canvas,crop);

sprite.Picture.Graphic := png2;

ReSetUp();

loader := false;

end;

end.
