unit uMain;
{************************************************************************
 * This program is  free software; you can redistribute it  and modify it
 * under the terms of the GNU  General Public License as published by the
 * Free Software Foundation; either version 2 of the license  or (at your
 * option) any later version.
 ***********************************************************************}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, DwmApi, IniFiles, pngimage, ottdimage, TileImage,
  ShellApi, StdCtrls, Buttons, ExtDlgs, ComCtrls, Math, Spin;

type
  TfrmSpriteEditor = class(TForm)
    aSpritePreview: TImage;
    PreviewContainer: TScrollBox;
    TileBackground: TTileImage;
    LoadSprite: TOpenPictureDialog;
    Panel1: TPanel;
    SpriteList: TListView;
    OTTDImage1: TOTTDImage;
    barTop: TBevel;
    Bevel1: TBevel;
    TemplateList: TComboBox;
    btnAddSprite: TSpeedButton;
    sPosition_x: TSpinEdit;
    sPosition_y: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    sOffset_x: TSpinEdit;
    sOffset_y: TSpinEdit;
    Label3: TLabel;
    Label4: TLabel;
    btnRemoveSprite: TSpeedButton;
    btnSaveSprites: TSpeedButton;
    btnCropActiveSprite: TSpeedButton;
    spriteHighlight: TShape;
    Timer1: TTimer;
    btnDetectOffset: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure ActivateSprite(Sender: TObject);
    procedure TemplateListChange(Sender: TObject);
    procedure btnAddSpriteClick(Sender: TObject);
    procedure SpriteListSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure ChangePreviewPosition(Sender: TObject);
    procedure btnRemoveSpriteClick(Sender: TObject);
    procedure ChangeSpriteOffset(Sender: TObject);
    procedure btnSaveSpritesClick(Sender: TObject);
    procedure btnCropActiveSpriteClick(Sender: TObject);
    procedure CloseSpriteHighlight(Sender: TObject);
    procedure btnDetectOffsetClick(Sender: TObject);
  private
    { Private declarations }
    templates_path:string;

    tile_width: integer;
    tile_height: integer;

    tile_offset_x: integer;
    tile_offset_y: integer;

    last_sprite_x:integer;
    last_sprite_y:integer;

    active_sprite: TOTTDImage;
    // used for sprite position changes - prevents sprite jumps when setting initial value
    loader: boolean;
  public
    { Public declarations }
    procedure AcceptFiles( var msg : TMessage ); message WM_DROPFILES;

    procedure EnableGlass(disable_glass:boolean);
    procedure LoadTemplates();
    procedure EnableTemplate(template_name:string);

    procedure ResetSprites;
    procedure RemoveActiveSprite();
    procedure AddSprite(spritefilename:string);
    function GetPNGOffset(png: TPngImage):TPoint;
    procedure SetActiveSprite;
    procedure SaveSprite(sprite:TOTTDImage);
  end;

var
  frmSpriteEditor: TfrmSpriteEditor;

implementation

{$R *.dfm}

uses uCrop;

procedure TfrmSpriteEditor.AcceptFiles( var msg : TMessage );
const cnMaxFileNameLen = 255;
var
  i, nCount  : integer;
  acFileName : array [0..cnMaxFileNameLen] of char;
begin
  // find out how many files we're accepting
  nCount := DragQueryFile( msg.WParam, $FFFFFFFF, acFileName, cnMaxFileNameLen );

  // query Windows one at a time for the file name
  for i := 0 to nCount-1 do
  begin
    DragQueryFile( msg.WParam, i, acFileName, cnMaxFileNameLen );
    // add the sprite
    if LowerCase(ExtractFileExt(acFileName))='.png' then AddSprite(acFileName);
  end;
  // let Windows know that you're done
  DragFinish( msg.WParam );
end;

procedure TfrmSpriteEditor.EnableGlass(disable_glass:boolean);
var dwm_active : longbool;
begin
dwm_active := false;
try
  DwmApi.DwmIsCompositionEnabled(dwm_active);
  if dwm_active and not disable_glass then GlassFrame.Enabled := true;
except end;
end;

procedure TfrmSpriteEditor.LoadTemplates;
var
  sr : TSearchRec;
begin
try
  if FindFirst(templates_path + '*.*', faDirectory, sr) < 0 then exit
  else
  repeat
    if ((sr.Attr and faDirectory <> 0) and (sr.Name <> '.') and (sr.Name <> '..')) then
      begin
      if FileExists(templates_path + sr.Name + '\template.cfg') then
         TemplateList.Items.Add(sr.Name);
      end;
  until FindNext(sr) <> 0;
finally
  SysUtils.FindClose(sr) ;
end;
end;


procedure TfrmSpriteEditor.EnableTemplate(template_name: string);
var ini:TIniFile;
    background:string;
begin
ResetSprites;

ini := TIniFile.Create(templates_path+template_name+'\template.cfg');

background := ini.ReadString('VPNGC_Template','BackgroundImage','plain.png');
TileBackground.Picture.LoadFromFile(templates_path+template_name+'\'+background);

tile_width := ini.ReadInteger('VPNGC_Template','tile_width', 32);
tile_height:= ini.ReadInteger('VPNGC_Template','tile_height',16);

tile_offset_x := ini.ReadInteger('VPNGC_Template', 'tile_offset_x', 0);
tile_offset_y := ini.ReadInteger('VPNGC_Template', 'tile_offset_y', 0);

ini.Free;

active_sprite := nil;

aSpritePreview.Picture := nil;
end;


procedure TfrmSpriteEditor.ResetSprites;
var i:integer;
begin
loader := true;
aSpritePreview.Picture := nil;
active_sprite := nil;
SpriteList.Items.Clear;

for i := (PreviewContainer.ComponentCount - 1) downto 0 do
  begin
    if PreviewContainer.Components[i] is TOTTDImage then
      begin
      try
        PreviewContainer.Components[i].Destroy;
      except
        end;
      end;
  end;

last_sprite_x := 0;
last_sprite_y := 2;


sPosition_x.Value := 0;
sPosition_y.Value := 0;
sOffset_x.Value := 0;
sOffset_y.Value := 0;

loader := false;
end;

procedure TfrmSpriteEditor.RemoveActiveSprite;
var i:integer;
    sl_id, s_id:integer;
begin
sl_id := SpriteList.ItemIndex;

aSpritePreview.Picture := nil;
active_sprite := nil;

for i := (PreviewContainer.ComponentCount - 1) downto 0 do
    if PreviewContainer.Components[i] is TOTTDImage then
      begin
      s_id := (PreviewContainer.Components[i] as TOTTDImage).Tag;
      if s_id = sl_id then
         begin
         // id matches. remove sprite
         PreviewContainer.Components[i].Destroy;
         SpriteList.Items.Delete(sl_id);
         end
      else
         begin
         // no id match. check to see if bigger than id, if yes, than substract 1
         if s_id > sl_id then
            (PreviewContainer.Components[i] as TOTTDImage).Tag := s_id-1;
         end;
      end;

loader := true;
sPosition_x.Value := 0;
sPosition_y.Value := 0;
sOffset_x.Value := 0;
sOffset_y.Value := 0;
loader := false;
end;


procedure TfrmSpriteEditor.AddSprite(spritefilename: string);
var newimg:TPngImage;
    img:TOTTDImage;
    pngOffset:TPoint;
    li:TListItem;
begin
loader := true;
last_sprite_x := last_sprite_x + 1;
last_sprite_y := last_sprite_y + 1;

if last_sprite_x > 5 then
  begin
    last_sprite_x := 1;
    last_sprite_x := 3;
  end;

newimg := TPngImage.Create;
newimg.LoadFromFile(spritefilename);

pngOffset := GetPNGOffset(newimg);

li:=SpriteList.Items.Add;
li.Caption := ChangeFileExt(ExtractFileName(spritefilename),'');
li.SubItems.Add(IntToStr(pngOffset.X));
li.SubItems.Add(IntToStr(pngOffset.Y));


Img := TOTTDImage.Create(PreviewContainer);
Img.Tag := li.Index;
Img.Parent:=PreviewContainer;
img.AutoSize:=true;
img.Picture.Assign(newimg);
img.SpriteOffsetX := pngOffset.X;
img.SpriteOffsetY := pngOffset.Y;
img.SpriteLocationX := last_sprite_x;
img.SpriteLocationY := last_sprite_y;
img.SpriteName := ExtractFileName(spritefilename);
Img.SpritePath := spritefilename;

Img.OnClick := ActivateSprite;

img.Left := (last_sprite_x * frmSpriteEditor.tile_width) + img.SpriteOffsetX + tile_offset_x;
img.Top  := (last_sprite_y * frmSpriteEditor.tile_height) + img.SpriteOffsetY + tile_offset_y;

img.BringToFront;

loader := false;
end;


// Get PNG offsets from the PNG file if they already are defined, else sets them to zero
function TfrmSpriteEditor.GetPNGOffset(png: TPngImage):TPoint;
var p:TPoint;
    i:integer;
    tEXt: TChunktEXt;
    found_x, found_y:boolean;
begin
p.X:=0;
p.Y:=0;
found_x := false;
found_y := false;
try
begin
//Memo1.Lines.Clear;
for i := (png.Chunks.Count - 1) downto 0 do
  begin
  if png.Chunks.Item[i].ClassType = TChunktEXt then
    begin
    tEXt := png.Chunks.Item[i] as TChunktEXt;
    if AnsiString(tEXt.Keyword)=AnsiString('x_offs') then
       begin
         p.X := StrToInt(trim(tEXt.Text));
         found_x := true;
         // remove the chunk from the memory copy
         png.Chunks.RemoveChunk(tEXt);
       end
    else
    if AnsiString(tEXt.Keyword)=AnsiString('y_offs') then
       begin
         p.Y := StrToInt(trim(tEXt.Text));
         found_y := true;
         // remove the chunk from the memory copy
         png.Chunks.RemoveChunk(tEXt);
       end;
    end
  end;
end
except
  // Something went wrong with the detection
  p.X := 0;  found_x := false;
  p.Y := 0;  found_y := false;
end;

// create offset from sprite width
//if not found_x then p.X := (floor(png.Width/2)-1)*(-1);
// create offset from sprite height
//if not found_y then p.Y := (png.Height - (tile_height*2))*(-1);
Result:=p;
end;

procedure TfrmSpriteEditor.SetActiveSprite;
begin
loader := true;
SpriteList.ItemIndex := active_sprite.Tag;

spriteHighlight.Left    := active_sprite.Left;
spriteHighlight.Top     := active_sprite.Top;
spriteHighlight.Width   := active_sprite.Width;
spriteHighlight.Height  := active_sprite.Height;
spriteHighlight.Visible := true;
spriteHighlight.BringToFront;

aSpritePreview.Picture.Assign(active_sprite.Picture);
if (active_sprite.Picture.Width > aSpritePreview.Width) or (active_sprite.Picture.Height > aSpritePreview.Height) then
  aSpritePreview.Stretch := true
else
  aSpritePreview.Stretch := false;

// preview position
sPosition_x.Value := active_sprite.SpriteLocationX;
sPosition_y.Value := active_sprite.SpriteLocationY;

// sprite offsets
sOffset_x.Value := active_sprite.SpriteOffsetX;
sOffset_y.Value := active_sprite.SpriteOffsetY;

Timer1.Enabled := false;
Timer1.Enabled := true;
loader := false;
end;


procedure TfrmSpriteEditor.SaveSprite(sprite:TOTTDImage);
{Characters for the PNG header}
const
  PngHeader: Array[0..7] of AnsiChar = (#137, #80, #78, #71, #13, #10, #26, #10);

var png:TPngImage;
    i,j:integer;
    saveto:TFileStream;
    p:TPoint;
begin
png:= sprite.Picture.Graphic as TPngImage;

// clear existing x_offs and y_offs is present
p:=GetPNGOffset(png);

{
compare original offsets and if they are the same,
do nothing and cancel further sprite processing
}
if (p.X = sprite.SpriteOffsetX) and (p.Y = sprite.SpriteOffsetY) then exit;

// add new offsets defined
png.AddtEXt('x_offs',IntToStr(sprite.SpriteOffsetX));
png.AddtEXt('y_offs',IntToStr(sprite.SpriteOffsetY));

{
we don't use the PngImage savetofile() function
because it writes the tEXt chunks at the end.
}

// initialize stream
saveto := TFileStream.Create(sprite.SpritePath, fmCreate);
{Reads the header}
saveto.Write(PNGHeader[0], 8);
// write TChunkIHDR chunk
for j := 0 to png.Chunks.Count - 1 do
  if png.Chunks.Item[j] is TChunkIHDR then
     png.Chunks.Item[j].SaveToStream(saveto);
// write TChunktEXt chunks
for j := 0 to png.Chunks.Count - 1 do
  if png.Chunks.Item[j] is TChunktEXt then
     png.Chunks.Item[j].SaveToStream(saveto);

// write all the other chunks
for j := 0 to png.Chunks.Count - 1 do
  if (not (png.Chunks.Item[j] is TChunkIHDR)) and (not (png.Chunks.Item[j] is TChunktEXt)) then
     png.Chunks.Item[j].SaveToStream(saveto);

saveto.Free;

end;
















procedure TfrmSpriteEditor.FormCreate(Sender: TObject);
var ini:TIniFile;
    disable_glass:boolean;
    default_template:string;
begin
// tell windows we accept file drops
DragAcceptFiles( Handle, True );

last_sprite_x := 0;
last_sprite_y := 2;        // we set the Y position so the sprites are not cut off

loader:=false;

templates_path := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName))+'templates\';

ini:=TIniFile.Create(ChangeFileExt(Application.ExeName,'.cfg'));

disable_glass := ini.ReadBool('VisualPNGcodec','DisableAero',false);
EnableGlass(disable_glass);

LoadTemplates;
default_template := ini.ReadString('VisualPNGcodec','DefaultTemplate','standard');


EnableTemplate(default_template);

ini.Free;
end;

procedure TfrmSpriteEditor.ActivateSprite(Sender: TObject);
begin
active_sprite := Sender as TOTTDImage;
SetActiveSprite();
end;


procedure TfrmSpriteEditor.TemplateListChange(Sender: TObject);
begin
EnableTemplate(TemplateList.Items.Strings[ TemplateList.ItemIndex ]);
end;


procedure TfrmSpriteEditor.btnAddSpriteClick(Sender: TObject);
var i:integer;
begin
if LoadSprite.Execute then
  begin
  for i := 0 to LoadSprite.Files.Count - 1 do
    AddSprite(LoadSprite.Files.Strings[i]);
  end;
end;



procedure TfrmSpriteEditor.btnCropActiveSpriteClick(Sender: TObject);
var cropForm:TfrmCropSprite;
    Res:TModalResult;
    pngOffset:TPoint;
begin
if active_sprite = nil then exit;
try
    cropForm := TfrmCropSprite.Create(nil);
    cropForm.sprite.Picture := active_sprite.Picture;

    cropForm.SetUp();

    Res := cropForm.ShowModal;
finally
    if Res = mrOk then
      begin
        // crop applied successfully
        active_sprite.Picture := cropForm.sprite.Picture;
        //reset sprite info and position
        loader := true;
        pngOffset := GetPNGOffset(active_sprite.Picture.Graphic as TPNGImage);
        active_sprite.SpriteOffsetX := pngOffset.X;
        active_sprite.SpriteOffsetY := pngOffset.Y;
        sOffset_x.Value := pngOffset.X;
        sOffset_y.Value := pngOffset.Y;
        SpriteList.Items[active_sprite.Tag].SubItems.Strings[0] := IntToStr(pngOffset.X);
        SpriteList.Items[active_sprite.Tag].SubItems.Strings[1] := IntToStr(pngOffset.Y);
        active_sprite.Left := (active_sprite.SpriteLocationX * frmSpriteEditor.tile_width) + active_sprite.SpriteOffsetX + tile_offset_x;
        active_sprite.Top  := (active_sprite.SpriteLocationY * frmSpriteEditor.tile_height) + active_sprite.SpriteOffsetY + tile_offset_y;
        loader := false;
      end;
    cropForm.Free;
end;
end;

procedure TfrmSpriteEditor.btnDetectOffsetClick(Sender: TObject);
var p:TPoint;
begin
// create offset from sprite width
p.X := (floor(active_sprite.Width/2)-1)*(-1);
// create offset from sprite height
p.Y := (active_sprite.Height - (tile_height*2))*(-1);
loader := true;
active_sprite.SpriteOffsetX := p.X;
active_sprite.SpriteOffsetY := p.Y;
sOffset_x.Value := p.X;
sOffset_y.Value := p.Y;
SpriteList.Items[active_sprite.Tag].SubItems.Strings[0] := IntToStr(p.X);
SpriteList.Items[active_sprite.Tag].SubItems.Strings[1] := IntToStr(p.Y);
active_sprite.Left := (active_sprite.SpriteLocationX * frmSpriteEditor.tile_width) + active_sprite.SpriteOffsetX + tile_offset_x;
active_sprite.Top  := (active_sprite.SpriteLocationY * frmSpriteEditor.tile_height) + active_sprite.SpriteOffsetY + tile_offset_y;
loader := false;

end;

procedure TfrmSpriteEditor.btnRemoveSpriteClick(Sender: TObject);
begin
RemoveActiveSprite();
end;

procedure TfrmSpriteEditor.btnSaveSpritesClick(Sender: TObject);
var i:integer;
begin
for i := (PreviewContainer.ComponentCount - 1) downto 0 do
  if PreviewContainer.Components[i] is TOTTDImage then
    begin
    active_sprite := (PreviewContainer.Components[i] as TOTTDImage);
    SaveSprite( active_sprite );
    end;
ShowMessage('All sprites saved!');
end;

procedure TfrmSpriteEditor.SpriteListSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var i:integer;
begin
if Selected and (Item <> nil) then
  begin
  for i := 0 to PreviewContainer.ComponentCount - 1 do
    if PreviewContainer.Components[i] is TOTTDImage then
      if (PreviewContainer.Components[i] as TOTTDImage).Tag = Item.Index then
        begin
        active_sprite := PreviewContainer.Components[i] as TOTTDImage;
        SetActiveSprite();
        end;
  end;
end;


procedure TfrmSpriteEditor.ChangePreviewPosition(Sender: TObject);
var old_pos_x, old_pos_y : Integer;
    sender_dir:string;
begin
if not (active_sprite is TOTTDImage) then exit;

if not loader then
  begin
  sender_dir := 'x';
  if (Sender is TSpinEdit) then
      if (Sender as TSpinEdit).Tag=1 then
          sender_dir := 'y'
      else
          sender_dir := 'x';


  // begin - prevent desync between sprites and tilemap
  old_pos_x := PreviewContainer.HorzScrollBar.Position;
  old_pos_y := PreviewContainer.VertScrollBar.Position;

  PreviewContainer.HorzScrollBar.Position:=0;
  PreviewContainer.VertScrollBar.Position:=0;

  // end - prevent desync between sprites and tilemap

  loader := true;
  if sender_dir = 'x' then
    begin
    // sender is for X direction
    if (sPosition_x.Value mod 2)=0 then
      sPosition_y.Value := sPosition_y.Value+1
    else
      sPosition_y.Value := sPosition_y.Value-1;
    end
  else
    begin
      // sender is for Y direction
    if (sPosition_y.Value mod 2)=0 then
      sPosition_x.Value := sPosition_x.Value+1
    else
      sPosition_x.Value := sPosition_x.Value-1;
    end;
  loader := false;

  active_sprite.SpriteLocationX := sPosition_x.Value;
  active_sprite.SpriteLocationY := sPosition_y.Value;

  active_sprite.Left := (active_sprite.SpriteLocationX * tile_width) + active_sprite.SpriteOffsetX + tile_offset_x;
  active_sprite.Top  := (active_sprite.SpriteLocationY * tile_height) + active_sprite.SpriteOffsetY + tile_offset_y;

  // restore original scroll position
  PreviewContainer.HorzScrollBar.Position:=old_pos_x;
  PreviewContainer.VertScrollBar.Position:=old_pos_y;
  end;
end;

procedure TfrmSpriteEditor.ChangeSpriteOffset(Sender: TObject);
var old_pos_x, old_pos_y : Integer;
begin
if not loader then
  begin
  // begin - prevent desync between sprites and tilemap
  old_pos_x := PreviewContainer.HorzScrollBar.Position;
  old_pos_y := PreviewContainer.VertScrollBar.Position;

  PreviewContainer.HorzScrollBar.Position:=0;
  PreviewContainer.VertScrollBar.Position:=0;

  // end - prevent desync between sprites and tilemap

  active_sprite.SpriteOffsetX := sOffset_x.Value;
  active_sprite.SpriteOffsetY := sOffset_y.Value;

  SpriteList.Items[active_sprite.Tag].SubItems.Strings[0] := IntToStr(active_sprite.SpriteOffsetX);
  SpriteList.Items[active_sprite.Tag].SubItems.Strings[1] := IntToStr(active_sprite.SpriteOffsetY);


  active_sprite.Left := (active_sprite.SpriteLocationX * tile_width) + active_sprite.SpriteOffsetX + tile_offset_x;
  active_sprite.Top  := (active_sprite.SpriteLocationY * tile_height) + active_sprite.SpriteOffsetY + tile_offset_y;

  // restore original scroll position
  PreviewContainer.HorzScrollBar.Position:=old_pos_x;
  PreviewContainer.VertScrollBar.Position:=old_pos_y;
  end;
end;

procedure TfrmSpriteEditor.CloseSpriteHighlight(Sender: TObject);
begin
Timer1.Enabled := false;
spriteHighlight.Visible := false;
end;

end.
